@@ -1,12 +1,15 @@
 r"""Mutation models."""
 
+import gctree.utils as utils
+
 from ete3 import TreeNode
 import numpy as np
 from scipy.stats import poisson
 import random
 import scipy
 from Bio.Seq import Seq
-from gctree.utils import hamming_distance
+import historydag as hdag
+from multiset import FrozenMultiset
 from typing import Tuple, List, Callable
 
 
@@ -62,20 +65,8 @@ class MutationModel:
     @staticmethod
     def _disambiguate(sequence):
         r"""generator of all possible nt sequences implied by a sequence
-        containing Ns."""
-        # find the first N nucleotide
-        N_index = sequence.find("N")
-        # if there is no N nucleotide, yield the input sequence
-        if N_index == -1:
-            yield sequence
-        else:
-            for n_replace in "ACGT":
-                # ooooh, recursion
-                # NOTE: in python3 we could simply use "yield from..." instead of this loop
-                for sequence_recurse in MutationModel._disambiguate(
-                    sequence[:N_index] + n_replace + sequence[N_index + 1 :]
-                ):
-                    yield sequence_recurse
+        containing ambiguous bases."""
+        return _sequence_disambiguations(sequence)
 
     def mutability(self, kmer: str) -> Tuple[np.float64, np.float64]:
         r"""Returns the mutability of a central base of :math:`k`-mer, along with
@@ -305,7 +296,9 @@ class MutationModel:
                                 leaf.sequence, lambda0=lambda0[0], frame=frame
                             )
                         child = TreeNode()
-                        child.dist = hamming_distance(mutated_sequence, leaf.sequence)
+                        child.dist = utils.hamming_distance(
+                            mutated_sequence, leaf.sequence
+                        )
                         child.add_feature("sequence", mutated_sequence)
                         child.add_feature("abundance", 0)
                         child.add_feature("terminated", False)
@@ -380,3 +373,156 @@ class MutationModel:
 
         # return the uncollapsed tree
         return tree
+
+
+def _sequence_disambiguations(sequence, _accum=""):
+    """Iterates through possible disambiguations of sequence, recursively.
+
+    Recursion-depth-limited by number of ambiguity codes in sequence,
+    not sequence length.
+    """
+    if sequence:
+        for index, base in enumerate(sequence):
+            if base in utils.bases:
+                _accum += base
+            else:
+                for newbase in utils.ambiguous_dna_values[base]:
+                    yield from _sequence_disambiguations(
+                        sequence[index + 1 :], _accum=(_accum + newbase)
+                    )
+                return
+    yield _accum
+
+
+def _mutability_dagfuncs(
+    *args, splits: List[int] = [], **kwargs
+) -> hdag.utils.AddFuncDict:
+    """Return functions for counting mutability parsimony on the history DAG.
+
+    Mutability parsimony of a tree is the sum over all edges in the tree
+    of mutability distances between parent and child node sequences.
+
+    The mutability distance from an ancestral sequence to a target sequence is the sum of
+    :math:`-log(mutability * p)` over all sites which do match, where :math:`mutability`
+    is the mutation frequency of the k-mer surrounding the mutated base
+    (in the ancestral sequence), and :math:`p` is the transition probability to the new
+    base in the target sequence. Notice that this so-called distance is not symmetric.
+
+    These functions may not be stable numerically, but we expect every tree to have
+    a unique mutability parsimony score for non-degenerate mutability models, so
+    so this shouldn't matter in practice.
+
+    Args:
+        splits: A list of indices at which sequences are concatenated, but are not
+            adjacent in the genome.
+        Other arguments are passed to :meth:`mutation_model.MutationModel` constructor.
+
+    Returns:
+        A :meth:`historydag.utils.AddFuncDict` which may be passed as keyword arguments
+        to :meth:`historydag.HistoryDag.weight_count`, :meth:`historydag.HistoryDag.trim_optimal_weight`,
+        or :meth:`historydag.HistoryDag.optimal_weight_annotate`
+        methods to trim or annotate a :meth:`historydag.HistoryDag` according to mutability model parsimony.
+        Weight format is ``float``.
+    """
+
+    mutation_model = MutationModel(*args, **kwargs)
+    dist = _mutability_distance(mutation_model, splits=splits)
+
+    @hdag.utils.access_field("label")
+    @hdag.utils.ignore_ualabel(0)
+    @hdag.utils.access_field("sequence")
+    def distance(seq1, seq2):
+        return dist(seq1, seq2)
+
+    return hdag.utils.AddFuncDict(
+        {"start_func": lambda n: 0, "edge_weight_func": distance, "accum_func": sum},
+        name="Mut. Pars.",
+    )
+
+
+def _mutability_distance_precursors(
+    mutation_model: MutationModel, splits: List[int] = []
+):
+    chunk_idxs = list(zip([0] + splits, splits + [None]))
+    # Caching could be moved to the MutationModel class instead.
+    context_model = mutation_model.context_model.copy()
+    k = mutation_model.k
+    h = k // 2
+    # Build all sequences with (when k=5) one or two Ns on either end
+    templates = [
+        ("N" * left, "N" * (k - left - right), "N" * right)
+        for left in range(h + 1)
+        for right in range(h + 1)
+        if left != 0 or right != 0
+    ]
+
+    kmers_to_compute = [
+        leftns + stub + rightns
+        for leftns, ambig_stub, rightns in templates
+        for stub in _sequence_disambiguations(ambig_stub)
+    ]
+    # Cache all these mutabilities in context_model also
+    context_model.update(
+        {kmer: mutation_model.mutability(kmer) for kmer in kmers_to_compute}
+    )
+
+    def add_ns(seq: str):
+        ns = "N" * h
+        chunks = [seq[start:end] for start, end in chunk_idxs]
+        return ns + ns.join(chunks) + ns
+
+    @utils._check_distance_arguments
+    def mutpairs(seq1: str, seq2: str):
+        seq1N = add_ns(seq1)
+        seq2N = add_ns(seq2)
+        mut_idxs = [
+            index
+            for index, (base1, base2) in enumerate(zip(seq1N, seq2N))
+            if base1 != base2
+        ]
+        return FrozenMultiset((seq1N[i - h : i + h + 1], seq2N[i]) for i in mut_idxs)
+
+    def sum_minus_logp(pairs: FrozenMultiset):
+        # I have chosen to multiply substitution rate for central base
+        # with rate of new base. Not sure if this is a good choice.
+        if pairs:
+            # for floating point behavior
+            pairs = sorted(pairs.items())
+            p_arr = [
+                mult
+                * (
+                    np.log(context_model[mer][0])
+                    + np.log(context_model[mer][1][newbase])
+                )
+                for (mer, newbase), mult in pairs
+            ]
+            return -sum(p_arr)
+        else:
+            return 0.0
+
+    return (mutpairs, sum_minus_logp)
+
+
+def _mutability_distance(mutation_model: MutationModel, splits=[]):
+    """Returns a fast distance function based on passed mutation_model.
+
+    First, caches computed mutabilities for k-mers with k // 2 N's on either end. This
+    is pretty fast for k=5, but the distance function should be created once
+    and reused.
+
+    The returned distance function sums :math:`-log(mutability * p)` over all
+    sites which do match between its two sequence arguments, where ``mutability``
+    is the mutation frequency of the k-mer surrounding the mutated base (in the
+    first sequence argument) and ``p`` is the transition probability to the new
+    base.
+
+    Note that, in particular, this function is not symmetric on its  arguments.
+    """
+    mutpairs, sum_minus_logp = _mutability_distance_precursors(
+        mutation_model, splits=splits
+    )
+
+    def distance(seq1, seq2):
+        return sum_minus_logp(mutpairs(seq1, seq2))
+
+    return distance
